use bevy::{app::AppExit, prelude::*};
use bevy_egui::{egui, EguiContext, EguiPlugin};
use bevy_prototype_lyon::prelude::{
    DrawMode, FillMode, FillOptions, GeometryBuilder, Path, PathBuilder, ShapePlugin, StrokeMode,
};

#[derive(Default, Resource)]
struct OccupiedScreenSpace {
    left: f32,
    top: f32,
    right: f32,
    bottom: f32,
}

#[derive(Default, Resource)]
struct Line {
    x: f32,
    y: f32,
}

#[derive(Resource, Deref, DerefMut)]
struct OriginalCameraTransform(Transform);

fn main() {
    App::new()
        .add_plugins(DefaultPlugins.set(WindowPlugin {
            window: WindowDescriptor {
                width: 700.0,
                height: 500.0,
                title: "Turtle Window".to_string(),
                present_mode: bevy::window::PresentMode::AutoVsync,
                //decorations: false,
                ..default()
            },
            ..default()
        }))
        .add_plugin(EguiPlugin)
        .add_plugin(ShapePlugin)
        .init_resource::<OccupiedScreenSpace>()
        .add_startup_system(setup_system)
        .add_system(ui)
        .add_system(update_path)
        .run();
}

fn update_path(line: Res<Line>, mut path: Query<&mut Path>) {
    for mut path in path.iter_mut() {
        let mut path_builder = PathBuilder::new();
        path_builder.move_to(Vec2::new(line.x, line.y));
        path_builder.line_to(Vec2::new(100., 0.));
        *path = path_builder.build();
    }
}

fn ui(
    mut egui_context: ResMut<EguiContext>,
    mut occupied_screen_space: ResMut<OccupiedScreenSpace>,
    mut line: ResMut<Line>,
    mut exit: EventWriter<AppExit>,
) {
    let mut style = (*egui_context.ctx_mut().style()).clone();
    style.visuals.button_frame = false;
    egui_context.ctx_mut().set_style(style);
    occupied_screen_space.left = 0.0;
    occupied_screen_space.right = egui::SidePanel::right("right_panel")
        .resizable(false)
        .show(egui_context.ctx_mut(), |ui| {
            ui.add_space(7.);
            ui.menu_button("Menu", |ui| {
                if ui
                    .add_sized([ui.available_width(), 30.], egui::Button::new("Exit"))
                    .clicked()
                {
                    exit.send(AppExit);
                }
            })
        })
        .response
        .rect
        .width();
    occupied_screen_space.top = 0.0;
    occupied_screen_space.bottom = egui::TopBottomPanel::bottom("bottom_panel")
        .resizable(false)
        .show(egui_context.ctx_mut(), |ui| {
            ui.add_sized(
                ui.available_size(),
                egui::Slider::new(&mut line.x, 0.0..=100.0),
            );
        })
        .response
        .rect
        .height();
}

fn setup_system(mut commands: Commands) {
    commands.spawn(Camera2dBundle::default());
    commands.insert_resource(Line { x: 100., y: 100. });
    let mut path_builder = PathBuilder::new();
    path_builder.move_to(Vec2::new(200., 200.));
    path_builder.line_to(Vec2::new(100., 0.));
    let line = path_builder.build();

    commands.spawn(GeometryBuilder::build_as(
        &line,
        DrawMode::Outlined {
            fill_mode: FillMode {
                options: FillOptions::default(),
                color: Default::default(),
            },
            outline_mode: StrokeMode::new(Color::BLACK, 6.),
        },
        Transform::IDENTITY,
    ));
}
