use bevy::prelude::{Color, Transform};
use bevy_prototype_lyon::prelude::PathBuilder;

use crate::{
    datatypes::{angle::Angle, length::Length, Coordinate, Speed, Visibility},
    structs::DrawingSegment,
    turtle::Precision,
};

/// Describing the full state of a turtle.
pub struct TurtleState {
    drawing: Vec<DrawingSegment>,
    position: Coordinate,
    heading: Angle<Precision>,
    color: Color,
    fill_color: Color,
    visible: Visibility,
    shape_transform: Transform,
    speed: Speed,
}

/// A default turtle drawing lines each line is one Segment.
pub struct Turtle {
    state: TurtleState,
}

/// A turtle that combines its commands to one closed shape with the `close()` command.
pub struct ShapeBuilder {
    state: TurtleState,
    builder: PathBuilder,
}

/// A turtle that draws a filled shape. End the filling process with the `end()` command.
pub struct FilledBuilder {
    state: TurtleState,
    builder: PathBuilder,
}

pub trait TurtleDraw {
    fn forward(&mut self, length: Length) -> &mut Self;
    fn backward(&mut self, length: Length) -> &mut Self;
    fn circle(&mut self, radius: Length, angle: Angle<Precision>) -> &mut Self;
    fn goto(&mut self, coordinate: Coordinate) -> &mut Self;
    fn home(&mut self) -> &mut Self;
    fn dot(&mut self, size: Length) -> &mut Self;
    fn stamp(&mut self, size: Length) -> &mut Self;
}

pub trait TurtleTurn {
    fn left(&mut self, angle: Angle<Precision>) -> &mut Self;
    fn right(&mut self, angle: Angle<Precision>) -> &mut Self;
    fn look_at(&mut self, coordinate: Coordinate) -> &mut Self;
}

pub trait TurtleHistory {
    fn reset(&mut self) -> &mut Self;
    fn clear(&mut self) -> &mut Self;
    fn undo(&mut self) -> &mut Self;

    fn clear_stamps(&mut self) -> &mut Self;
}
